/*
	mem.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2006 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/
#include "ngs_toolkit.h"


struct query_table* init_query_hash(struct query_table* qt,struct parameters* param)
{
	int i,j;
	qt = malloc(sizeof(struct query_table));
	assert(qt != 0);
	qt->total_expression = 0.0;
	qt->numseq = 0;
	qt->size = 0;
	qt->tree = 0;
	qt->tree_size = 0;
	qt->string = malloc(sizeof(char)*param->query_num_limit*30);
	assert(qt->string != 0);
	qt->hash = malloc(sizeof(struct query*)* (1<< (KEY_LEN+7)));
	assert(qt->hash != 0);
	for(i = 0;i < (1<< (KEY_LEN+7));i++){
		qt->hash[i] = 0;
	}
	qt->park = malloc(sizeof(struct query*)* param->query_num_limit);
	assert(qt->park != 0);
	for(i = 0; i < param->query_num_limit;i++){
		qt->park[i] = malloc(sizeof(struct query));
		assert(qt->park[i] != 0);
	}
	qt->stats = malloc(sizeof(double*) * 6);
	assert(qt->stats != 0);
	
	//if(!param->upper_limit){
		for (i = 0; i < 6;i++){
			qt->stats[i] = malloc(sizeof (double)*5);
			for(j = 0; j < 5;j++){
				qt->stats[i][j] = 0.0;
			}
		}
	/*}else if(param->upper_limit == 1){
		for (i = 0; i < 6;i++){
			qt->stats[i] = malloc(sizeof (double)*3);
			for(j = 0; j < 3;j++){
				qt->stats[i][j] = 0.0;
			}
		}
	}else{
		for (i = 0; i < 6;i++){
			qt->stats[i] = malloc(sizeof (double)*4);
			for(j = 0; j < 4;j++){
				qt->stats[i][j] = 0.0;
			}
		}
	}*/
	
	return qt;
}

void free_query_hash(struct query_table* qt,struct parameters* param)
{
	int i;
	for(i = 0; i < param->query_num_limit;i++){
		free(qt->park[i]);
	}
	for (i = 0; i < 6;i++){
		free(qt->stats[i]);
	}
	free(qt->stats);
	if(qt->tree){
		free_p_tree(qt->tree);
	}
	free(qt->park);
	free(qt->hash);
	free(qt->string);
	free(qt);
}

struct fasta* alloc_target(struct fasta* p,char** infiles,int num_infiles)
{
	int fd = 0;
	int i = 0;
	int test;

	int mer_hash = 0;
	int version = 0;
	int max_numseq = 0;
	int max_string_len = 0;
	
	p = malloc(sizeof(struct fasta));
	assert(p != 0);
	
	for(i = 0; i < num_infiles;i++){
		if ((fd = open(infiles[i], O_RDONLY )) == -1){
			
			//fprintf(stderr,"Cannot open file %s\n",infiles[i]);
			//exit(1);
		}else{
		version = 0;
		test = read(fd,&version,sizeof(int));
		assert(test >= 0);
		
		if(version != SUFFIX_VERSION){
			fprintf(stderr,"Warning: There has been a change to the format of the suffix files in the recent version. You need to re-build all suffix files.\n");
			exit(0);
			free(p);
		}
		
		test = read(fd,&mer_hash,sizeof(int));
		assert(test>=0);
		test = read(fd,&p->string_len,sizeof(int));
		if(p->string_len > max_string_len){
			max_string_len = p->string_len;
		}
		test = read(fd,&p->numseq,sizeof(int));
		if(p->numseq > max_numseq){
			max_numseq = p->numseq;
		}
		close(fd);
		}
	}
	
	//allocing... 
	
	//fprintf(stderr,"MAX:%d	%d\n",max_string_len,max_numseq);
		
	p->sn= malloc(sizeof(char*)*max_numseq);	
	assert(p->sn != 0);
	p->s_index= malloc(sizeof(int)*(max_numseq+1));	
	assert(p->s_index != 0);
	p->string = malloc(sizeof(char)*(max_string_len+1+17));
	assert(p->string != 0);

	p->suffix = malloc(sizeof(int)*(max_string_len+1));
	assert(p->suffix != 0);
	
	//p->mer_hash = malloc(sizeof(int)*(134217728));
	//assert(p->mer_hash != 0);
	return p;
}


void free_names(struct fasta*aln)
{
	int i;
	for (i =0;i < aln->numseq;i++){
		free(aln->sn[i]);
	//	free(aln->seq[i]);
	}
	free(aln->mer_hash);
	/*if(aln->suffix){
		free(aln->suffix);
	}
	
	free(aln->mer_hash);
	//if(aln->org_seq){
	//	free(aln->org_seq);
	//}
	free(aln->s_index);
	free(aln->string);
	free(aln->sn);
	//free(aln->seq);
	//free(aln->sl);
	
	//free(aln->e);
	free(aln);*/
}

void free_aln(struct fasta*aln)
{
	int i;
	for (i =0;i < aln->numseq;i++){
		free(aln->sn[i]);
	}
	free(aln->mer_hash);
	if(aln->suffix){
		free(aln->suffix);
	}
	
	//free(aln->mer_hash);
	free(aln->s_index);
	free(aln->string);
	free(aln->sn);

	free(aln);
}


void free_param(struct parameters* param)
{
	int i;
	
	if(param->types){
		for(i = 0;i < param->num_types;i++){
			free(param->types[i]);
		}
		free(param->types);
	}
	
	for(i=0;i< param->infiles;i++){
		free(param->suffixfile[i]);
		free(param->solid_suffixfile[i]);
	}
	
	if(param->infile){
		free(param->suffixfile);
		free(param->solid_suffixfile);
	}
	
	free(param->outfile_tmp);
	free(param->infile);
	free(param);
}
