/*
	pattern_searching.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2006 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"
#include "pattern_searching_solid.h"
#include "mapping_output.h"


#ifdef thread

void* pattern_matching_solid_thread(void *threadarg)
{
	
	struct thread_data *data;
	data = (struct thread_data *) threadarg;
	
	const struct query_table* qt = data->q;
	const struct fasta* t = data->t;
	
	const unsigned int upper_limit = data->param->upper_limit;
	const int start = data->start;
	const int end = data->end;
	const int format = data->param->format;
	const int mode = data->param->mode;
	struct query** tmp = qt->park;
	
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	//struct suffix_pos* sp_tmp = malloc(sizeof(struct suffix_pos));
	//assert(sp_tmp != 0);
	//char* mod_line = malloc(sizeof(char)*(20));
	//assert(mod_line != 0);
	int i;
	//int ii;
	sp->mis = 0;
	sp->indel = 0;
	
	if(mode & 1){
		for(i = start; i < end;i++){
			//len = q->s_index[i+1]-(q->s_index[i]+1);
			sp = pm_solid(sp,t,tmp[i]->seq,tmp[i]->len);
			tmp[i]->stats[0] += sp->sum;
			tmp[i]->stats[1] += sp->sum;
			if(sp->sum > 0 && tmp[i]->stats[0] <= upper_limit){
				mapping_output(sp,t,tmp[i],"E",format);
				
				
				/*sp_tmp->s_m = 0;
				sp_tmp->e_m = 0;
				for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
					sprintf(mod_line, "E,%d",base_space_errors(tmp[i]->seq,t->org_seq+t->suffix[ii]-1,tmp[i]->len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,sn,tmp[i]->len,mod_line,format);
				}
				sp_tmp->s_p = 0;
				sp_tmp->e_p = 0;
				for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
					sprintf(mod_line, "E,%d",base_space_errors_rev(tmp[i]->seq,t->org_seq+t->suffix[ii]-1,tmp[i]->len));
					sp_tmp->s_m = ii-1;
					sp_tmp->e_m= ii;
					mapping_output(sp_tmp,t,sn,tmp[i]->len,mod_line,format);
				}
				//mapping_output(sp,t,q->sn[i],len-1,"E",format);*/
			}
		}
	}
	if(mode & 2){
		for(i = start; i < end;i++){
			//if(tmp[i]->stats[0]  <= upper_limit){
				pm_solid_n_mis(t,tmp[i],tmp[i]->seq,upper_limit,format,0,data->param->mismatches,0,data->param->mismatches+1);
			//}
		}
	}
	/*if(mode & 4){
		for(i = start; i < end;i++){
			len = q->s_index[i+1]-(q->s_index[i]+1);
			sp->sum = pm_ins_solid(t,q->seq[i],q->sn[i],len,upper_limit,format);
			if(sp->sum){
				q->c[i][0] += sp->sum;
				q->c[i][5] += sp->sum;
				
			}
			//if(sp->sum){
			//	q->c[i] |= 0x20000000; 
			//	q->c[i]+= sp->sum;
			//}
			sp->sum = pm_del_solid(t,q->seq[i],q->sn[i],len,upper_limit,format);
			if(sp->sum){
				q->c[i][0] += sp->sum;
				q->c[i][5] += sp->sum;	
			}
			//if(sp->sum){
			//	q->c[i] |= 0x20000000; 
			//	q->c[i]+= sp->sum;
			//}
		}
	}*/
	/*
	if(mode & 8){
		for(i = start; i < end;i++){
			len = q->s_index[i+1]-(q->s_index[i]+1);
			sp->sum = pm_mis_misl1(t,q->seq[i],q->sn[i],len,upper_limit,format);
			if(sp->sum){
				q->c[i] |= 0x10000000; 
				q->c[i]+= sp->sum;
			}
		}
	}*/
	//free(mod_line);
	//free(sp_tmp);
	free(sp);
	pthread_exit((void *) 0);
}

void* pattern_matching_solid_quick_thread(void *threadarg)
{
	
	struct thread_data *data;
	data = (struct thread_data *) threadarg;
	
	const struct query_table* qt = data->q;
	const struct fasta* t = data->t;
	
	const unsigned int upper_limit = data->param->upper_limit;
	const int start = data->start;
	const int end = data->end;
	const int format = data->param->format;
	const int mode = data->param->mode;
	struct query** tmp = qt->park;
	
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	//struct suffix_pos* sp_tmp = malloc(sizeof(struct suffix_pos));
	//assert(sp_tmp != 0);
	//char* mod_line = malloc(sizeof(char)*(20));
	//assert(mod_line != 0);
	int i;
	//int ii;
	sp->mis = 0;
	sp->indel = 0;
	
	if(mode & 1){
		for(i = start; i < end;i++){
			if(tmp[i]->stats[0]  <= upper_limit){
				sp = pm_solid(sp,t,tmp[i]->seq,tmp[i]->len);
				tmp[i]->stats[0] += sp->sum;
				//tmp[i]->stats[1] += sp->sum;
				if(sp->sum > 0 && tmp[i]->stats[0] <= upper_limit){
					mapping_output(sp,t,tmp[i],"E",format);
				}
			}
		}
	}
	if(mode & 2){
		for(i = start; i < end;i++){
			if(tmp[i]->stats[0]  <= upper_limit){
				pm_solid_n_mis_quick(t,tmp[i],tmp[i]->seq,upper_limit,format,0,data->param->mismatches,0,data->param->mismatches+1);
			}
		}
	}
	free(sp);
	pthread_exit((void *) 0);
}


#else

void pattern_matching_solid(struct query_table* qt,const struct fasta* t,const struct parameters* param)
{
	struct query** tmp = qt->park;
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	//struct suffix_pos* sp_tmp = malloc(sizeof(struct suffix_pos));
	//assert(sp_tmp != 0);
	
	const unsigned int upper_limit = param->upper_limit;
	const int  format = param->format;
	const int mode = param->mode;
	char* mod_line = malloc(sizeof(char)*(20));
	assert(mod_line != 0);
	int i,a;
	//int ii;

	sp->mis = 0;
	sp->indel = 0;
	a = 1;
	if(mode & 1){
		for(i = 0; i < qt->size;i++){
			//len = q->s_index[i+1]-(q->s_index[i]+1);
			sp = pm_solid(sp,t,tmp[i]->seq,tmp[i]->len);
			tmp[i]->stats[0] += sp->sum;
			tmp[i]->stats[1] += sp->sum;
			if(sp->sum > 0 && tmp[i]->stats[0] <= upper_limit){
				mapping_output(sp,t,tmp[i],"E",format);
				
				/*sp_tmp->s_m = 0;
				sp_tmp->e_m = 0;
				for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
					sprintf(mod_line, "E,%d",base_space_errors(tmp[i]->seq,t->org_seq+t->suffix[ii]-1,tmp[i]->len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,sn,tmp[i]->len,mod_line,format);
				}
				sp_tmp->s_p = 0;
				sp_tmp->e_p = 0;
				for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
					sprintf(mod_line, "E,%d",base_space_errors_rev(tmp[i]->seq,t->org_seq+t->suffix[ii]-1,tmp[i]->len));
					sp_tmp->s_m = ii-1;
					sp_tmp->e_m = ii;
					mapping_output(sp_tmp,t,sn,tmp[i]->len,mod_line,format);
				}*/
				//mapping_output(sp,t,q->sn[i],len-1,"E",format);
			}
		}
	}
	if(mode & 2){
		for(i = 0; i < qt->size;i++){
			pm_solid_n_mis(t,tmp[i],tmp[i]->seq,upper_limit,format,0,param->mismatches,0,param->mismatches+1);
		}
	}
	free(mod_line);
	//free(sp_tmp);
	free(sp);
}

void pattern_matching_solid_quick(struct query_table* qt,const struct fasta* t,const struct parameters* param)
{
	struct query** tmp = qt->park;
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	//struct suffix_pos* sp_tmp = malloc(sizeof(struct suffix_pos));
	//assert(sp_tmp != 0);
	
	const unsigned int upper_limit = param->upper_limit;
	const int  format = param->format;
	const int mode = param->mode;
	char* mod_line = malloc(sizeof(char)*(20));
	assert(mod_line != 0);
	int i,a;
	//int ii;

	sp->mis = 0;
	sp->indel = 0;
	a = 1;
	if(mode & 1){
		for(i = 0; i < qt->size;i++){
			if(tmp[i]->stats[0]  <= upper_limit){
				sp = pm_solid(sp,t,tmp[i]->seq,tmp[i]->len);
				tmp[i]->stats[0] += sp->sum;
				tmp[i]->stats[1] += sp->sum;
				if(sp->sum > 0 && tmp[i]->stats[0] <= upper_limit){
					mapping_output(sp,t,tmp[i],"E",format);
				}
			}
		}
	}
	if(mode & 2){
		for(i = 0; i < qt->size;i++){
			if(tmp[i]->stats[0]  <= upper_limit){
				pm_solid_n_mis_quick(t,tmp[i],tmp[i]->seq,upper_limit,format,0,param->mismatches,0,param->mismatches+1);
			}
		}
	}
	free(mod_line);
	//free(sp_tmp);
	free(sp);
}
#endif





struct suffix_pos* pm_solid(struct suffix_pos* sp,const struct fasta*t ,const char *p,const int len)
{
	char rc[MAX_LINE];
	register unsigned int i = 0;
	register unsigned int r = 0;
	register int j;
	reverse(p,rc,len);
	//i = tag_index(p);
	//i = tag_index_solid(p);
	sp->e_p = 0;
	sp->s_p = 0;
	sp->e_m = 0;
	sp->s_m = 0;
	sp->sum = 0;
	for(j = 0; j < 16;j++){
		i = i << 2 ;
		i |= p[j] - 48;
		r = r << 2;
		r |= rc[j] - 48;
	}
	
	if(t->mer_hash[i>>5] & (1<<(i & 0x1F))){
		sp->e_p = binsearch_up(p,t->string,t->suffix,len,t->string_len);
		sp->s_p = binsearch_down(p,t->string,t->suffix,len,sp->e_p);
		sp->sum = (sp->e_p-sp->s_p);
	}
	//i = tag_index(rc);
	//i = tag_index_solid(rc);
	if(t->mer_hash[r>>5] & (1<<(r & 0x1F))){
		sp->e_m = binsearch_up(rc,t->string,t->suffix,len,t->string_len);
		sp->s_m = binsearch_down(rc,t->string,t->suffix,len,sp->e_m);
		sp->sum +=(sp->e_m-sp->s_m);
	}
	return sp;
}


void pm_solid_n_mis(const struct fasta*t ,struct query* q,const char *p,const unsigned int upper_limit,const int format,int start,int n,char* mod_line, int mis)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	//assert(sp != 0);
	char* alphabet = "0123";
	char tmp_seq[MAX_LINE];// = 0;
	int m_col[5] = {-1,-1,2,4,5};
	int i,j,c;
	
	int mod_len = 0;
	
	sp->mis = mis-n;
	sp->indel = 0;

	if(!mod_line){
		sp->indel = 1;
		mod_line = malloc(sizeof(char)*(10*n));
		assert(mod_line != 0);
	}else{
		mod_len = strlen(mod_line);
		mod_line[mod_len] = ',';
		mod_len++;
	}
	
	n--;
	c = m_col[mis-n];

	for(j =0; j < q->len;j++){
		tmp_seq[j] = p[j];
	}
	tmp_seq[j] = 0;
	for(i = start;i < q->len;i++){
		if(i == 16){
			sp = pm_solid(sp,t,tmp_seq,16);
			if(!sp->sum){
				break;
			}
		}
		for(j = 0;j < 4;j++){
			if(alphabet[j] != p[i]){
				tmp_seq[i] = alphabet[j];
				sp = pm_solid(sp,t,tmp_seq,q->len);
				
				q->stats[c] += sp->sum;
				q->stats[0] += sp->sum;
				if(sp->sum > 0 && q->stats[0] <= upper_limit){
					sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
					mapping_output(sp,t,q,mod_line,format);
				}
				if(n){
					sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
					pm_solid_n_mis(t,q,tmp_seq,upper_limit,format,i+1,n,mod_line,mis);
				}
			}
		}
		tmp_seq[i] = p[i];
	}
	if(sp->indel){
		free(mod_line);
	}
	//free(tmp_seq);
	//free(sp_tmp);
	free(sp);
}

void pm_solid_n_mis_quick(const struct fasta*t ,struct query* q,const char *p,const unsigned int upper_limit,const int format,int start,int n,char* mod_line, int mis)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	//assert(sp != 0);
	char* alphabet = "0123";
	char tmp_seq[MAX_LINE];// = 0;
	//int m_col[5] = {-1,-1,2,4,5};
	int i,j;//,c;
	
	int mod_len = 0;
	
	sp->mis = mis-n;
	sp->indel = 0;

	if(!mod_line){
		sp->indel = 1;
		mod_line = malloc(sizeof(char)*(10*n));
		assert(mod_line != 0);
	}else{
		mod_len = strlen(mod_line);
		mod_line[mod_len] = ',';
		mod_len++;
	}
	
	n--;
	//c = m_col[mis-n];
	/*if(start < 2){
		start = 2;
	}*/

	for(j =0; j < q->len;j++){
		tmp_seq[j] = p[j];
	}
	tmp_seq[j] = 0;
	for(i = start;i < q->len;i++){
		if(i == 16){
			sp = pm_solid(sp,t,tmp_seq,16);
			if(!sp->sum){
				break;
			}
		}
		for(j = 0;j < 4;j++){
			if(alphabet[j] != p[i]){
				tmp_seq[i] = alphabet[j];
				sp = pm_solid(sp,t,tmp_seq,q->len);
				
				//q->stats[c] += sp->sum;
				q->stats[0] += sp->sum;
				if(q->stats[0]  > upper_limit){
					goto endloop;
				}
				if(sp->sum > 0 && q->stats[0] <= upper_limit){
					sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
					mapping_output(sp,t,q,mod_line,format);
				}
				if(n){
					sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
					pm_solid_n_mis(t,q,tmp_seq,upper_limit,format,i+1,n,mod_line,mis);
				}
			}
		}
		tmp_seq[i] = p[i];
	}
	
endloop:
		
	
	if(sp->indel){
		free(mod_line);
	}
	//free(tmp_seq);
	//free(sp_tmp);
	free(sp);
}


/*
int pm_solid_mis(const struct fasta*t ,const char *p,const char* q_name,const int len,const unsigned int upper_limit,const int format)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	struct suffix_pos* sp_tmp = malloc(sizeof(struct suffix_pos));
	assert(sp_tmp != 0);
	char* alphabet = "0123";
	char* tmp_seq = 0;
	char* mod_line = 0;
	int i,j,f,g;
	int sum = 0;
	int used_cc[4] = {0,0,0,0};
	int a = 0;
	int b = 0;
	int ii;
	
	sp->mis = 1;
	sp->indel = 0;
	tmp_seq = malloc(sizeof(char)*(len+1));
	assert(tmp_seq);
	mod_line = malloc(sizeof(char)*(20));
	assert(mod_line);
	
	for(i = 2;i < len;i++){
		g = 0;
		for(f = 0; f < i;f++){
			tmp_seq[g] = p[f];
			g++;
		}
		for(j = 0;j < 4;j++){
			if(alphabet[j] != p[i]){
				g = i;
				tmp_seq[g] = alphabet[j];
				g++;
				
				for(f = i+1;f < len;f++){
					tmp_seq[g] = p[f];
					g++;
				}
				tmp_seq[g] = 0;
				//fprintf(stderr,"%s	%s\n",p,tmp_seq);
				sp = pm_solid(sp,t,tmp_seq+2,len-2);
				if(sp->sum > 0 && sp->sum <= upper_limit){
					sp_tmp->s_m = 0;
					sp_tmp->e_m = 0;
					for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
						sprintf(mod_line, "M%d:%c:%c,%d",i,p[i],alphabet[j],base_space_errors(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
						sp_tmp->s_p = ii-1;
						sp_tmp->e_p = ii;
						mapping_output(sp_tmp,t,q_name,len,mod_line,format);
					}
					sp_tmp->s_p = 0;
					sp_tmp->e_p = 0;
					for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
						sprintf(mod_line, "M%d:%c:%c,%d",i,p[i],alphabet[j],base_space_errors_rev(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
						sp_tmp->s_p = ii-1;
						sp_tmp->e_p = ii;
						mapping_output(sp_tmp,t,q_name,len,mod_line,format);
					}
					//sprintf(mod_line, "SM%d",i);
					//mapping_output(sp,t,q_name,len,mod_line,format);
				}
				sum += sp->sum;
			}
		}
	}
	
	//fprintf(stderr,"\n\n");
	for(i = 2;i < len-1;i++){
		g = 0;
		for(f = 0; f < i;f++){
			tmp_seq[g] = p[f];
			g++;
		}
		used_cc[0] = 0;
		used_cc[1] = 0;
		used_cc[2] = 0;
		used_cc[3] = 0;
		
		if(p[i] == p[i+1]){
			used_cc[(int)(p[i]-48)] = 1;
			for(j = 0;j < 4;j++){
				if(!used_cc[j]){
					g = i;
					tmp_seq[g] = alphabet[j];
					g++;
					tmp_seq[g] = alphabet[j];
					g++;
					
					for(f = i+2;f < len;f++){
						tmp_seq[g] = p[f];
						g++;
					}
					tmp_seq[g] = 0;
					//fprintf(stderr,"%s	%s\n",p,tmp_seq);
					sp = pm_solid(sp,t,tmp_seq+2,len-2);
					if(sp->sum > 0 && sp->sum <= upper_limit){
						sp_tmp->s_m = 0;
						sp_tmp->e_m = 0;
						for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
							sprintf(mod_line, "GM%d,%d",i,base_space_errors(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
							sp_tmp->s_p = ii-1;
							sp_tmp->e_p = ii;
							mapping_output(sp_tmp,t,q_name,len,mod_line,format);
						}
						sp_tmp->s_p = 0;
						sp_tmp->e_p = 0;
						for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
							sprintf(mod_line, "GM%d,%d",i,base_space_errors_rev(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
							sp_tmp->s_p = ii-1;
							sp_tmp->e_p = ii;
							mapping_output(sp_tmp,t,q_name,len,mod_line,format);
						}
					}
					sum += sp->sum;
				}
			}
		}else{
			used_cc[(int)(p[i]-48)] = 1;
			used_cc[(int)(p[i+1]-48)] = 1;
			g = i;
			tmp_seq[g] = alphabet[(int)(p[i+1]-48)];
			g++;
			tmp_seq[g] = alphabet[(int)(p[i]-48)];
			g++;
			
			for(f = i+2;f < len;f++){
				tmp_seq[g] = p[f];
				g++;
			}
			tmp_seq[g] = 0;
			//fprintf(stderr,"%s	%s\n",p,tmp_seq);
			sp = pm_solid(sp,t,tmp_seq+2,len-2);
			if(sp->sum > 0 && sp->sum <= upper_limit){
				sp_tmp->s_m = 0;
				sp_tmp->e_m = 0;
				for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
					sprintf(mod_line, "GM%d,%d",i,base_space_errors(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,q_name,len,mod_line,format);
				}
				sp_tmp->s_p = 0;
				sp_tmp->e_p = 0;
				for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
					sprintf(mod_line, "GM%d,%d",i,base_space_errors_rev(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,q_name,len,mod_line,format);
				}
			}
			sum += sp->sum;
			
			
			for(j = 0;j < 4;j++){
				if(!used_cc[j]){
					a = j;
					used_cc[j] = 1;
					break;
				}
			}
			for(j = 0;j < 4;j++){
				if(!used_cc[j]){
					b = j;
				}
			}
			g = i;
			tmp_seq[g] = alphabet[a];
			g++;
			tmp_seq[g] = alphabet[b];
			g++;
			
			for(f = i+2;f < len;f++){
				tmp_seq[g] = p[f];
				g++;
			}
			tmp_seq[g] = 0;
			//fprintf(stderr,"%s	%s\n",p,tmp_seq);
			sp = pm_solid(sp,t,tmp_seq+2,len-2);
			if(sp->sum > 0 && sp->sum <= upper_limit){
				sp_tmp->s_m = 0;
				sp_tmp->e_m = 0;
				for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
					sprintf(mod_line, "GM%d,%d",i,base_space_errors(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,q_name,len,mod_line,format);
				}
				sp_tmp->s_p = 0;
				sp_tmp->e_p = 0;
				for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
					sprintf(mod_line, "GM%d,%d",i,base_space_errors_rev(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,q_name,len,mod_line,format);
				}
			}
			sum += sp->sum;
			
			
			g = i;
			tmp_seq[g] = alphabet[b];
			g++;
			tmp_seq[g] = alphabet[a];
			g++;
			
			for(f = i+2;f < len;f++){
				tmp_seq[g] = p[f];
				g++;
			}
			tmp_seq[g] = 0;
			//fprintf(stderr,"%s	%s\n",p,tmp_seq);
			sp = pm_solid(sp,t,tmp_seq+2,len-2);
			if(sp->sum > 0 && sp->sum <= upper_limit){
				sp_tmp->s_m = 0;
				sp_tmp->e_m = 0;
				for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
					sprintf(mod_line, "GM%d,%d",i,base_space_errors(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,q_name,len,mod_line,format);
				}
				sp_tmp->s_p = 0;
				sp_tmp->e_p = 0;
				for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
					sprintf(mod_line, "GM%d,%d",i,base_space_errors_rev(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,q_name,len,mod_line,format);
				}
			}
			sum += sp->sum;
			
		}
	}
	
	free(tmp_seq);
	free(mod_line);
	free(sp);
	free(sp_tmp);
	return sum;
}


int pm_del_solid(const struct fasta*t,const char *p,const char* q_name,const int len,const unsigned int upper_limit,const int format)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	struct suffix_pos* sp_tmp = malloc(sizeof(struct suffix_pos));
	assert(sp_tmp != 0);
	char* tmp_seq = 0;
	char* mod_line = 0;
	int i,j,g;
	int ii;
	int sum = 0;
	sp->mis = 0;
	sp->indel = 1;
	tmp_seq = malloc(sizeof(char)*(len));
	assert(tmp_seq != 0);
	mod_line = malloc(sizeof(char)*(20));
	assert(mod_line != 0);
	
	for(i = 1;i < len-1;i++){
		//es[c] = malloc(sizeof(char)*(len));
		//es[c+1] = malloc(sizeof(char)*(20));
		if(p[i] != p[i+1]){
			g = 0;// used to be 0
			for(j = 0;j < len;j++){//used to be 0
				if(j != i){
					tmp_seq[g] = p[j];
					g++;
				}
			}
			tmp_seq[g] = 0;
			
			sp = pm_solid(sp,t,tmp_seq+2,len-3);
			if(sp->sum > 0 && sp->sum <= upper_limit){
				sp_tmp->s_m = 0;
				sp_tmp->e_m = 0;
				for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
					sprintf(mod_line, "D%d,%d",i,base_space_errors(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,q_name,len,mod_line,format);
				}
				sp_tmp->s_p = 0;
				sp_tmp->e_p = 0;
				for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
					sprintf(mod_line, "D%d,%d",i,base_space_errors_rev(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
					sp_tmp->s_p = ii-1;
					sp_tmp->e_p = ii;
					mapping_output(sp_tmp,t,q_name,len,mod_line,format);
				}
			}
			sum += sp->sum;
		}
	}
	g = 0;
	for(j = 0;j < len-1;j++){
		tmp_seq[g] = p[j];
		g++;
		
	}
	tmp_seq[g] = 0;
	
	sp = pm_solid(sp,t,tmp_seq+2,len-3);
	if(sp->sum > 0 && sp->sum <= upper_limit){
		sp_tmp->s_m = 0;
		sp_tmp->e_m = 0;
		for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
			sprintf(mod_line, "D%d,%d",i,base_space_errors(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
			sp_tmp->s_p = ii-1;
			sp_tmp->e_p = ii;
			mapping_output(sp_tmp,t,q_name,len,mod_line,format);
		}
		sp_tmp->s_p = 0;
		sp_tmp->e_p = 0;
		for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
			sprintf(mod_line, "D%d,%d",i,base_space_errors_rev(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
			sp_tmp->s_p = ii-1;
			sp_tmp->e_p = ii;
			mapping_output(sp_tmp,t,q_name,len,mod_line,format);
		}
	}
	sum += sp->sum;
	
	free(tmp_seq);
	free(mod_line);
	free(sp_tmp);
	free(sp);
	return sum;
}


int pm_ins_solid(const struct fasta*t ,const char *p,const char* q_name,const int len,const unsigned int upper_limit,const int format)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp != 0);
	struct suffix_pos* sp_tmp = malloc(sizeof(struct suffix_pos));
	assert(sp_tmp != 0);
	char* alphabet = "0123";
	char* tmp_seq = 0;
	char* mod_line = 0;
	int i,j,f,g;
	int sum = 0;
	sp->mis = 0;
	sp->indel = 1;
	tmp_seq = malloc(sizeof(char)*(len+2));
	assert(tmp_seq != 0);
	mod_line = malloc(sizeof(char)*(20));
	assert(mod_line != 0);
	
	for(i = 1;i < (len-2);i++){
		for(j = 0;j < 4;j++){
			if(p[i+1] != alphabet[j]){
				g = 0;
				for(f = 0; f <= i;f++){
					tmp_seq[g] = p[f];
					g++;
				}
				tmp_seq[g] = alphabet[j];
				g++;
				
				for(f = i+1;f < len;f++){
					tmp_seq[g] = p[f];
					g++;
				}
				tmp_seq[g] = 0;
				
				sp = pm_solid(sp,t,tmp_seq+2,len-1);
				if(sp->sum > 0 && sp->sum <= upper_limit){
					sprintf(mod_line, "I%d%c",i,alphabet[j]);
					mapping_output(sp,t,q_name,len,mod_line,format);
				}
				sum += sp->sum;
			}
		}
	}
	i = len-2;
	for(j = 0;j < 4;j++){
		g = 0;
		for(f = 0; f <= i;f++){
			tmp_seq[g] = p[f];
			g++;
		}
		tmp_seq[g] = alphabet[j];
		g++;
		
		for(f = i+1;f < len;f++){
			tmp_seq[g] = p[f];
			g++;
		}
		tmp_seq[g] = 0;
		
		
		sp = pm_solid(sp,t,tmp_seq+2,len-1);
		if(sp->sum > 0 && sp->sum <= upper_limit){
			sprintf(mod_line, "I%d%c",i,alphabet[j]);
			mapping_output(sp,t,q_name,len,mod_line,format);
		}
		sum += sp->sum;
	}
	
	free(tmp_seq);
	free(mod_line);
	free(sp_tmp);
	free(sp);
	return sum;
}

int pm_solid_n_mis(const struct fasta*t ,const char *p,const char* q_name,unsigned int* c,const int len,const unsigned int upper_limit,const int format,int start,int n,char* mod_line, int mis)
{
	struct suffix_pos* sp = malloc(sizeof(struct suffix_pos));
	assert(sp);
	struct suffix_pos* sp_tmp = malloc(sizeof(struct suffix_pos));
	assert(sp_tmp);
	char* alphabet = "0123";
	char* tmp_seq = 0;
	int i,j,f,g;
	int sum = 0;
	
	int ii;
	int mod_len = 0;
	
	sp->mis = 1;
	sp->indel = 0;
	
	tmp_seq = malloc(sizeof(char)*(len+1));
	assert(tmp_seq != 0);
	if(!mod_line){
		sp->indel = 1;
		mod_line = malloc(sizeof(char)*(10*n));
		assert(mod_line != 0);
		mod_len = 0;
	}else{
		mod_len = strlen(mod_line);
		mod_line[mod_len] = ',';
		mod_len++;
	}
	
	n--;
	for(i = start+2;i < len;i++){
		g = 0;
		for(f = 0; f < i;f++){
			tmp_seq[g] = p[f];
			g++;
		}
		for(j = 0;j < 4;j++){
			if(alphabet[j] != p[i]){
				g = i;
				tmp_seq[g] = alphabet[j];
				g++;
				
				for(f = i+1;f < len;f++){
					tmp_seq[g] = p[f];
					g++;
				}
				tmp_seq[g] = 0;
				//sprintf(mod_line+mod_len, "M%d:%c:%c",i,p[i],alphabet[j]);
				//fprintf(stderr,"%s	%s	%s	%d\n",p,tmp_seq,mod_line,n);
				sp = pm_solid(sp,t,tmp_seq+2,len-2);
				if(sp->sum > 0 && sp->sum <= upper_limit){
					sp_tmp->s_m = 0;
					sp_tmp->e_m = 0;
					for(ii = sp->s_p+1;ii <= sp->e_p; ii++){
						sprintf(mod_line+mod_len, "M%d:%c:%c,%d",i,p[i],alphabet[j],base_space_errors(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
						sp_tmp->s_p = ii-1;
						sp_tmp->e_p = ii;
						mapping_output(sp_tmp,t,q_name,len,mod_line,format);
					}
					sp_tmp->s_p = 0;
					sp_tmp->e_p = 0;
					for(ii = sp->s_m+1;ii <= sp->e_m; ii++){
						sprintf(mod_line+mod_len, "M%d:%c:%c,%d",i,p[i],alphabet[j],base_space_errors_rev(tmp_seq,t->org_seq+t->suffix[ii]-1,len));
						sp_tmp->s_p = ii-1;
						sp_tmp->e_p = ii;
						mapping_output(sp_tmp,t,q_name,len,mod_line,format);
					}
					//sprintf(mod_line, "SM%d",i);
					//mapping_output(sp,t,q_name,len,mod_line,format);
				}
				c[mis-n] += sp->sum;
				sum += sp->sum;
				//sum += sp->sum;
				if(n){
					sum += pm_solid_n_mis(t,tmp_seq,q_name,c,len,upper_limit,format,i-1,n,mod_line,mis);
				}
				
			}
		}
	}
	if(sp->indel){
		free(mod_line);
	}
	free(tmp_seq);
	free(sp_tmp);
	free(sp);
	
	return sum;
}
*/

