/*
 *  r_output.c
 *  NGSnexalign
 *
 *  Created by Timo Lassmann on 02/05/2008.
 *  Copyright 2008 RIKEN Yokohama Institute, Genome Exploration Research Group. All rights reserved.
 *
 */

#include "ngs_toolkit.h"

void q_sort(double* cat_expression,char** cat_name,char** cat_color, int left, int right);
void quickSort(double* cat_expression,char** cat_name,char** cat_color, int array_size);

void quickSort(double* cat_expression,char** cat_name,char** cat_color, int array_size)
{
	q_sort(cat_expression,cat_name,cat_color, 0, array_size - 1);
}

void q_sort(double* cat_expression,char** cat_name,char** cat_color, int left, int right)
{
	double pivot;
	int l_hold, r_hold;
	char* pivot2;
	char* pivot3;
	l_hold = left;
	r_hold = right;
	
	pivot = cat_expression[left];
	pivot2 = cat_name[left];
	pivot3 = cat_color[left];
	while (left < right){
		while ((cat_expression[right]>= pivot) && (left < right)){
			right--;
		}
		if (left != right){
			cat_expression[left] = cat_expression[right];
			cat_name[left] = cat_name[right];
			cat_color[left] = cat_color[right];
			left++;
		}
		while ((cat_expression[left] <= pivot) && (left < right)){
			left++;
		}
		if (left != right){
			cat_expression[right] = cat_expression[left];
			cat_name[right] = cat_name[left];
			cat_color[right] = cat_color[left];
			right--;
		}
	}
	cat_expression[left] = pivot;
	cat_name[left] = pivot2;
	cat_color[left] = pivot3;
	pivot = left;
	left = l_hold;
	right = r_hold;
	
	if (left < pivot){
		q_sort(cat_expression,cat_name,cat_color, left, pivot-1);
	}
	
	if (right > pivot){
		q_sort(cat_expression,cat_name,cat_color, pivot+1, right);
	}
}

void r_ann_pie(struct query_table* qt,struct parameters* param)
{
	char** cat_name = 0;
	char** cat_color = 0;
	char* filename = 0;
	double* cat_expression = 0;
	FILE *fout = 0;
	char* basename = 0; 
	int i,j;
	int pic;
	int pid;
	
	int num_cat = 1;
	
	cat_expression = malloc(sizeof(double)* (qt->tree_size+1));
	assert(cat_expression != 0);
	cat_name = malloc(sizeof(char*) * (qt->tree_size+1));
	assert(cat_name != 0);
	cat_color = malloc(sizeof(char*) * (qt->tree_size+1));
	assert(cat_color !=0);
	
	cat_color[0] = "black";
	cat_name[0] = malloc(sizeof(char)*6);
	cat_name[0][0] = 'O';
	cat_name[0][1] = 'T';
	cat_name[0][2] = 'H';
	cat_name[0][3] = 'E';
	cat_name[0][4] = 'R';
	cat_name[0][5] = 0;
	
	cat_expression[0] = 0.0;
	
	if(param->types){
		num_cat = collect_pie_data(qt->tree,param->types,param->type_colors,cat_expression,cat_name,cat_color,1,qt->total_expression);
	}else{
		num_cat = collect_pie_data(qt->tree,param->infile,0,cat_expression,cat_name,cat_color,1,qt->total_expression);
	}
	
	//for(i = 0;i < num_cat;i++){
	//	fprintf(stderr,"%s	%f	%s\n",cat_name[i],cat_expression[i],cat_color[i]);
	//}
	
	quickSort(cat_expression,cat_name,cat_color,num_cat);
	
	//for(i = 0;i < num_cat;i++){
	//	fprintf(stderr,"%s	%f	%s\n",cat_name[i],cat_expression[i],cat_color[i]);
	//}
	
	if(!strcmp(param->ann,"STDOUT")){
		param->ann = "annotation";
	}
	basename = malloc(sizeof(char)*strlen(param->ann));
	assert(basename);
	
	j = 0;
	for (i = strlen(param->ann)-1;i > 0; i--){
		if(param->ann[i] == '.'){
			j = i;
			break;
		}else if(param->ann[i] == '/'){
			break;
		}
	}
	if(j ==0){
		j = strlen(param->ann);
	}

	filename = malloc(sizeof(char)*(j+3));
	assert(filename != 0);
	for(i = 0;i < j;i++){
		basename[i] = param->ann[i];
		filename[i] = param->ann[i];
	}
	basename[i] = 0;
	filename[i+0] = '.';
	filename[i+1] = 'R';
	filename[i+2] = 0;
	

	assert(filename != 0);
	
	
	
	if ((fout = fopen(filename, "w")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
		
	fprintf(fout,"basefilename = paste(\"%s\");\n",basename);
	fprintf(fout,"ps_title = paste(\"%s\",\": \",\"%0.0f Sequences\",sep = '');\n",shorten_name(basename),qt->total_expression);
	fprintf(fout,"pngfilename = paste(basefilename,\".png\",sep = '');\n");
	fprintf(fout,"jpegfilename = paste(basefilename,\".jpeg\",sep = '');\n");
	fprintf(fout,"pdffilename = paste(basefilename,\".pdf\",sep = '');\n");
				
	fprintf(fout,"x = c(");
	if(cat_expression[0]){
		fprintf(fout,"%f,",cat_expression[0]);
	}
	for(i = 1; i < num_cat-1;i++){
		fprintf(fout,"%f,",cat_expression[i]);
	}
	fprintf(fout,"%f);\n",cat_expression[i]);
	
	fprintf(fout,"names = c(");
	if(cat_expression[0]){
		fprintf(fout,"paste(\"%s\",\" \",\"(\",\"%0.2f%%\",\")\",sep=''),",cat_name[0],cat_expression[0]);
	}
	for(i = 1; i < num_cat-1;i++){
		fprintf(fout,"paste(\"%s\",\" \",\"(\",\"%0.2f%%\",\")\",sep=''),",cat_name[i],cat_expression[i]);
	}
	fprintf(fout,"paste(\"%s\",\" \",\"(\",\"%0.2f%%\",\")\",sep=''));\n",cat_name[i],cat_expression[i]);
	
	fprintf(fout,"colors = c(");
	for(i = 0; i < num_cat-1;i++){
		fprintf(fout,"\"%s\",",cat_color[i]);
	}
	fprintf(fout,"\"%s\");\n",cat_color[i]);
	
	

	for(pic = 0; pic < 3;pic++){
		switch (pic){
			case 0:
				//fprintf(fout,"postscript(file=epsfilename,width=12.0,height=12.0,horizontal=F,paper=\"special\",family=\"Helvetica\",onefile = FALSE,pointsize=%d,bg =\"white\",title=ps_title)\n",24);
				
				fprintf(fout,"pdf(file=pdffilename,width=12.0,height=12.0,paper=\"special\",onefile = FALSE,pointsize=%d,bg=\"white\",title=ps_title,family = \"Helvetica\")\n",24);
				
				break;
			case 1:
				//fprintf(fout,"jpeg(file=jpegfilename,width=1200.0,height=1200.0,unit = \"px\",horizontal=F,paper=\"special\",family=\"Helvetica\",onefile = FALSE,pointsize=%d,bg =\"white\",quality = 100)\n",24);
				
				
				fprintf(fout,"jpeg(file=jpegfilename,width=1200.0,height=1200.0,quality = 100,bg=\"white\",pointsize=%d)\n",24);
				fprintf(fout,"par(family = \"sans\");\n");
				
				break;
			case 2:
				fprintf(fout,"png(file=pngfilename,width=1200.0,height=1200.0,bg=\"white\",pointsize=%d)\n",24);
				fprintf(fout,"par(family = \"sans\");\n");
				//fprintf(fout,"png(file=pngfilename,width=1200.0,height=1200.0,unit = \"px\",horizontal=F,paper=\"special\",family=\"Helvetica\",onefile = FALSE,pointsize=%d,bg =\"white\",quality = 100)\n",24);
				break;
		}
		fprintf(fout,"pie(x,names,col = colors,main = ps_title);\n");
		fprintf(fout,"dev.off()\n");
	}
	fclose(fout);
	
	pid = fork();
	//printf("The process id is %d and return value is %d\n",	getpid(), return_value);
	if ( pid == 0 ){ 
		execl("/usr/bin/env","R","R","CMD","BATCH",filename,"logfile.out",NULL);
	}else{
		wait(0);
		/*if ((unlink(filename)) == -1){
			perror("Removing temporary file failed");
			exit(1);
		}*/
	}
}

void r_stats_pie(struct query_table* qt,struct parameters* param)
{
	//char *names[] = {"OTHER","Unmapped","Single","1-10","11-100",">100"};
	char* colors[] = {"red","darkgreen","green","lightgreen","blue"};
	FILE *fout = 0;
	char* basename = 0; 
	int i,j;
	int pic;
	int pid;
	int last = 5;
	char* filename = 0;		
	char names[6][MAX_LINE];
	
	
	if(!strcmp(param->stats,"STDOUT")){
		param->stats = "mapping_stats";
	}
	
	basename = malloc(sizeof(char)*strlen(param->stats));
	assert(basename);
	j = 0;
	for (i = strlen(param->stats)-1;i > 0; i--){
		if(param->stats[i] == '.'){
			j = i;
			break;
		}else if(param->stats[i] == '/'){
			break;
		}
	}
	if(j ==0){
		j = strlen(param->stats);
	}
	
	filename = malloc(sizeof(char)*(j+3));
	for(i = 0;i < j;i++){
		basename[i] = param->stats[i];
		filename[i] = param->stats[i];
	}
	basename[i] = 0;
	filename[i+0] = '.';
	filename[i+1] = 'R';
	filename[i+2] = 0;
	
	
	assert(filename != 0);
	
	
	
	if ((fout = fopen(filename, "w")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	
	if(!param->upper_limit){
		for(i = 0; i < 5;i++){
			qt->stats[0][i] = qt->stats[0][i]/qt->total_expression*100;
			/*if(qt->stats[0][i] < 1.0){
				other_exp += qt->stats[0][i];
			}else{
				last = i;
			}*/
			//fprintf(stderr,"%f	%d\n",qt->stats[0][i],last);
		}
		last = 4;
		sprintf(names[0], "Unmapped");
		sprintf(names[1], "Single");
		sprintf(names[2], "2-10");
		sprintf(names[3], "11-100");
		sprintf(names[4], ">100");
		
	}else if(param->upper_limit == 1){
		for(i = 0; i < 3;i++){
			qt->stats[0][i] = qt->stats[0][i]/qt->total_expression*100;
			/*if(qt->stats[0][i] < 1.0){
				other_exp += qt->stats[0][i];
			}else{
				last = i;
			}*/
		}
		sprintf(names[0], "Unmapped");
		sprintf(names[1], "Single");
		sprintf(names[2], ">1");
		//sprintf(names[4], "11-100");
		//sprintf(names[5], ">100");
		last = 2;
		
	}else{
		for(i = 0; i < 4;i++){
			qt->stats[0][i] = qt->stats[0][i]/qt->total_expression*100;
			/*if(qt->stats[0][i] < 10.0){
				other_exp += qt->stats[0][i];
			}else{
				last = i;
			}*/
		}	
		last = 3;
		sprintf(names[0], "Unmapped");
		sprintf(names[1], "Single");
		sprintf(names[2], "2-%d",param->upper_limit);
		sprintf(names[3], ">%d",param->upper_limit);
		//sprintf(names[5], ">100");
	}
	
	
	
	
	
	fprintf(fout,"basefilename = paste(\"%s\");\n",basename);
	fprintf(fout,"ps_title = paste(\"%s\",\": \",\"%0.0f Sequences\",sep = '');\n",shorten_name(basename),qt->total_expression);
	fprintf(fout,"pngfilename = paste(basefilename,\".png\",sep = '');\n");
	fprintf(fout,"jpegfilename = paste(basefilename,\".jpeg\",sep = '');\n");
	fprintf(fout,"pdffilename = paste(basefilename,\".pdf\",sep = '');\n");
	
	fprintf(fout,"x = c(");
	//if(other_exp){
	//	fprintf(fout,"%f,",other_exp);
	//}
	for(i = 0; i < last;i++){
		//if(qt->stats[0][i] >= 1.0){
		fprintf(fout,"%f,",qt->stats[0][i]);
		//}
	}
	fprintf(fout,"%f);\n",qt->stats[0][last]);
	
	
	fprintf(fout,"names = c(");
	//if(other_exp){
	//	fprintf(fout,"\"%s\",",names[0]);
	//}
	
	for(i = 0; i < last;i++){
		//if(qt->stats[0][i] >= 1.0){
			fprintf(fout,"paste(\"%s\",\" \",\"(\",\"%0.2f%%\",\")\",sep=''),",names[i],qt->stats[0][i]);
		//}
	}
	fprintf(fout,"paste(\"%s\",\" \",\"(\",\"%0.2f%%\",\")\",sep=''));\n",names[last],qt->stats[0][last]);
	
	
	fprintf(fout,"colors = c(");
	//if(other_exp){
	//	fprintf(fout,"\"%s\",",colors[0]);
	//}
	
	for(i = 0; i < last;i++){
	//	if(qt->stats[0][i] >= 1.0){
			fprintf(fout,"\"%s\",",colors[i]);
	//	}
	}
	fprintf(fout,"\"%s\");\n",colors[last]);
	
		
	for(pic = 0; pic < 3;pic++){
		switch (pic){
			case 0:
				//fprintf(fout,"postscript(file=epsfilename,width=12.0,height=12.0,horizontal=F,paper=\"special\",family=\"Helvetica\",onefile = FALSE,pointsize=%d,bg =\"white\",title=ps_title)\n",24);
				
				fprintf(fout,"pdf(file=pdffilename,width=12.0,height=12.0,paper=\"special\",onefile = FALSE,pointsize=%d,bg=\"white\",title=ps_title,family = \"Helvetica\")\n",24);
				
				
				break;
			case 1:
				//fprintf(fout,"jpeg(file=jpegfilename,width=1200.0,height=1200.0,unit = \"px\",horizontal=F,paper=\"special\",family=\"Helvetica\",onefile = FALSE,pointsize=%d,bg =\"white\",quality = 100)\n",24);
				
				fprintf(fout,"jpeg(file=jpegfilename,width=1200.0,height=1200.0,quality = 100,bg=\"white\",pointsize=%d)\n",24);
				fprintf(fout,"par(family = \"sans\");\n");
				break;
			case 2:
				//fprintf(fout,"png(file=pngfilename,width=1200.0,height=1200.0,unit = \"px\",horizontal=F,paper=\"special\",family=\"Helvetica\",onefile = FALSE,pointsize=%d,bg =\"white\",quality = 100)\n",24);
				
				fprintf(fout,"png(file=pngfilename,width=1200.0,height=1200.0,bg=\"white\",pointsize=%d)\n",24);
				fprintf(fout,"par(family = \"sans\");\n");

				break;
		}

		fprintf(fout,"pie(x,names,col = colors,main = ps_title);\n");
		fprintf(fout,"dev.off()\n");
	}
	fclose(fout);
	
	pid = fork();
	//printf("The process id is %d and return value is %d\n",	getpid(), return_value);
	if ( pid == 0 ){ 
		//execl("/usr/bin/R","R","CMD","BATCH",filename,"logfile.out",NULL);
		execl("/usr/bin/env","R","R","CMD","BATCH",filename,"logfile.out",NULL);
		
	}else{
		wait(0);
		/*if ((unlink(filename)) == -1){
		 perror("Removing temporary file failed");
		 exit(1);
		 }*/
	}
}

