/*
	sort_mapping.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2007,8 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"
#include "mapping.h"

/*void sort_mapping_output(struct parameters* param)
{
	
}*/

static int cmpmap(const void *a, const void *b);

void sort_db(struct parameters* param)
{
	struct mapping** map = 0;
	struct sort_node* sort_tree = 0;
	FILE *file = 0;
	FILE *foutfile = 0;
	char *time_s = 0;
	char* sort_target = 0;

	char test1[MAX_LINE];
	
	char test2[MAX_LINE];
	char test3[MAX_LINE];
	char test4[MAX_LINE];
	float tag_count = 0.0;
	int map_num = 0;
	int start,stop,i;
	int cont = 1;
	int lbound = 0;
	int ubound = JUMP_SIZE;
	int count = 0;
	int min_jump = 0;
	
	if(!param->quiet_flag){
		time_s = get_time(time_s);
		fprintf(stderr,"%s	Sorting Output\n",time_s);
	}
	
	if (!(file = fopen(param->outfile, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->outfile);
		exit(-1);
	}
	
	if ((foutfile = fopen(param->outfile_tmp, "w")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	

	
	while(cont){
		if(cont == 2){
			lbound += JUMP_SIZE+min_jump;
			ubound += JUMP_SIZE+min_jump;
			if(!param->quiet_flag){
				fprintf(stderr,"Sorting:	%s	+	%d	%d\n",sort_target,lbound,ubound);
			}
		}else{
			lbound = 0;
			ubound = JUMP_SIZE;
			free(sort_target);
			sort_target = 0;
			
		}
		
		
		min_jump = INT_MAX;
		count = 0;
		cont = 0;
		while(fscanf(file, "%s%s%s%d%d%s%f%d", test1,test2,test3,&start,&stop,test4,&tag_count,&map_num)!= EOF){
			//fprintf(stderr,"ECHO %s",line);
			if(test3[0] == '+'){
 			if(!sort_target && !exists_sort_node(sort_tree,test2)){
				sort_target = strdup(test2);
				//fprintf(stderr,"Sorting:	%s	+	%d	%d\n",sort_target,lbound,ubound);
				sort_tree = insert_sort_node(sort_tree,test2);
			}
			if(sort_target){
				if(!strcmp(sort_target,test2)){
					if(start > ubound){
						if(start - ubound - 1< min_jump){
							min_jump = start - ubound - 1;
						}
						cont = 2;
					}else if(lbound <=  start){
						map = (struct mapping**)realloc(map, (count + 1) * sizeof(struct mapping*));
						map[count] = (struct mapping *)malloc(sizeof(struct mapping));
						map[count]->qname = strdup(test1);
						map[count]->tname = strdup(test2);
						map[count]->strand = strdup(test3);
						map[count]->edit = strdup(test4);
						map[count]->start = start;
						map[count]->stop = stop;
						map[count]->tag_count = tag_count;
						map[count]->map_num = map_num;
						//sort_tree = insert_sort_node(sort_tree,map[count]->tname);
						count++;
					}
				}else{
					if(!cont){
						cont = 1;
					}	
				}
			}
			/*fprintf(stderr,"1:%s	",test1);
			fprintf(stderr,"2:%s	",test2);
			fprintf(stderr,"3:%s	",test3);
			fprintf(stderr,"4:%d	",start);
			fprintf(stderr,"5:%d	",stop);
			fprintf(stderr,"6:%s\n",test4);*/
			}
		}
		if(!count){
			cont = 0;
		}else{
			//fprintf(stderr,"Sorting:%s	%s	%d-%d	%d	%d\n",sort_target,"+",lbound,ubound,count,min_jump);
			//printf("Before:\n");
			//for(i = 0; i < count; i++) {
			//	fprintf(stderr,"%s	%s	%s	%d	%d	%s\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit);
			//}

			/* qsort array of structures */
			qsort(map, count, sizeof(*map), cmpmap);

			/* after sorting ... */
			//printf("\n--\nAfter:\n");
			for(i = 0; i < count; i++) {
				fprintf(foutfile,"%s	%s	%s	%d	%d	%s	%f	%d\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit,map[i]->tag_count,map[i]->map_num);
			}		
			for(i = 0; i < count; i++) {
				free(map[i]->qname);
				free(map[i]->tname);
				free(map[i]->strand);
				free(map[i]->edit);
				free(map[i]);
			}
			free(map);
			map = 0;
			rewind(file);
		}
	}
		
	rewind(file);
	if(sort_tree){
		free_sort_tree(sort_tree);
		sort_tree = 0;
	}
	
	
	cont = 1;
	lbound = 0;
	ubound = JUMP_SIZE;
	count = 0;
	min_jump = 0;
	while(cont){
		if(cont == 2){
			lbound += JUMP_SIZE+min_jump;
			ubound += JUMP_SIZE+min_jump;
			if(param->quiet_flag){
				fprintf(stderr,"Sorting:	%s	-	%d	%d\n",sort_target,lbound,ubound);
			}
		}else{
			lbound = 0;
			ubound = JUMP_SIZE;
			free(sort_target);
			sort_target = 0;
		}
		min_jump = INT_MAX;
		count = 0;
		cont = 0;
		while(fscanf(file, "%s%s%s%d%d%s%f%d", test1,test2,test3,&start,&stop,test4,&tag_count,&map_num)!= EOF){
			if(test3[0] == '-'){
			//fprintf(stderr,"ECHO %s",line);
			if(!sort_target && !exists_sort_node(sort_tree,test2)){
				sort_target = strdup(test2);
//fprintf(stderr,"Sorting:	%s	-	%d	%d\n",sort_target,lbound,ubound);
				sort_tree = insert_sort_node(sort_tree,test2);
			}
			if(sort_target){
				if(!strcmp(sort_target,test2)){
					if(start > ubound){
						if(start - ubound - 1< min_jump){
							min_jump = start - ubound - 1;
						}
						cont = 2;
					}else if(lbound <=  start){
						map = (struct mapping**)realloc(map, (count + 1) * sizeof(struct mapping*));
						map[count] = (struct mapping *)malloc(sizeof(struct mapping));
						map[count]->qname = strdup(test1);
						map[count]->tname = strdup(test2);
						map[count]->strand = strdup(test3);
						map[count]->edit = strdup(test4);
						map[count]->start = start;
						map[count]->stop = stop;
						map[count]->tag_count = tag_count;
						map[count]->map_num = map_num;
						//sort_tree = insert_sort_node(sort_tree,map[count]->tname);
						count++;
					}
				}else{
					if(!cont){
						cont = 1;
					}	
				}
			}
			/*fprintf(stderr,"1:%s	",test1);
			fprintf(stderr,"2:%s	",test2);
			fprintf(stderr,"3:%s	",test3);
			fprintf(stderr,"4:%d	",start);
			fprintf(stderr,"5:%d	",stop);
			fprintf(stderr,"6:%s\n",test4);*/
			}
		}
		if(!count){
			cont = 0;
		}else{
			//fprintf(stderr,"Sorting:%s	%s	%d-%d	%d	%d\n",sort_target,"-",lbound,ubound,count,min_jump);
			//printf("Before:\n");
			//for(i = 0; i < count; i++) {
			//	fprintf(stderr,"%s	%s	%s	%d	%d	%s\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit);
			//}

			/* qsort array of structures */
			qsort(map, count, sizeof(*map), cmpmap);

			/* after sorting ... */
			//printf("\n--\nAfter:\n");
			for(i = 0; i < count; i++) {
				fprintf(foutfile,"%s	%s	%s	%d	%d	%s	%f	%d\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit,map[i]->tag_count,map[i]->map_num);
			}	
			for(i = 0; i < count; i++) {
				free(map[i]->qname);
				free(map[i]->tname);
				free(map[i]->strand);
				free(map[i]->edit);
				free(map[i]);
			}
			free(map);
			map = 0;
			rewind(file);
		}
	}
	
	
	
	fclose(file);
	fclose(foutfile);
	
	if(sort_tree){
		free_sort_tree(sort_tree);
	}
	
	
	if (!(file = fopen(param->outfile_tmp, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->outfile);
		exit(-1);
	}
	
	if ((foutfile = fopen(param->outfile, "w")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	
	while(fgets(test1, MAX_LINE, file)){
		fprintf(foutfile,"%s",test1);
	}
	fclose(foutfile);
	fclose(file);
	
	if(!param->quiet_flag){
		time_s = get_time(time_s);
		fprintf(stderr,"%s	Done\n",time_s);
	}
	
}


static int cmpmap(const void *a, const void *b)
{ 
	struct mapping * const *one = a;
	struct mapping * const *two = b;
	return (*one)->start -  (*two)->start;
}




struct sort_node* insert_sort_node(struct sort_node *n,char* name)
{
	int i;
	if(n == 0){
		n = malloc(sizeof(struct sort_node));
		assert(n != 0);
		n->left = 0;
		n->right = 0;
		n->name = strdup(name);
	}else{
		i = strcmp(n->name,name);
				
		if(i < 0){
			n->left = insert_sort_node(n->left,name);
		}else if(i > 0){
			n->right = insert_sort_node(n->right,name);
		}else if(i == 0){
			//n->count += count;
			return n;
		}
	}
	return n;
}

int exists_sort_node(struct sort_node *n,char* name)
{
	int i;
	if(n == 0){
		return 0;
	}else{
		i = strcmp(n->name,name);
				
		if(i < 0){
			return exists_sort_node(n->left,name);
		}else if(i > 0){
			return exists_sort_node(n->right,name);
		}else if(i == 0){
			//n->count += count;
			return 1;
		}
	}
	return 0;
}



void free_sort_tree(struct sort_node* n)
{
	if(n->left){
		free_sort_tree(n->left);
	}
	if(n->right){
		free_sort_tree(n->right);
	}
	free(n->name);
	free(n);
}






